-- $Author: gvozden $
-- $Date: 2004/01/22 15:16:07 $
-- $Revision: 1.2 $

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;	   
use ieee.std_logic_arith.all;	 

package wb_test_suite is  
	 
---------------------------------------------------------------------- 
-- type declarations 
---------------------------------------------------------------------- 
	type wb_cycle_data is 
		record
			value     : std_logic_vector(31 downto 0); 
			exp_value : std_logic_vector(31 downto 0); 
			data_width: natural;
			address   : std_logic_vector(31 downto 0);
			adr_width : natural;
			wr        : boolean;
		end record wb_cycle_data;

	type wb_cycle_data_array is array (NATURAL range <>) of wb_cycle_data; 
	
	type wb_status_type is (OK, ERROR, RETRY);
											
---------------------------------------------------------------------- 
-- function declarations 
---------------------------------------------------------------------- 
 	function log2(
	 	A: integer
	) return integer;
	
	function vector_to_string (
		I: in std_logic_vector
	) return string;
	
---------------------------------------------------------------------- 
-- procedure declarations 
---------------------------------------------------------------------- 

	-- purpose: upis podataka u WB slave
	-- usage  : unutar procesa
	procedure wbs_wr(
		constant data     : in  std_logic_vector;
		constant address  : in  std_logic_vector;
		signal   clk      : in  std_logic;
		signal   stb      : out std_logic;
		signal   ack      : in  std_logic;
		signal   err      : in  std_logic;
		signal   rty      : in  std_logic;
		signal   sel      : out std_logic_vector( 3 downto 0);
		signal   dat      : out std_logic_vector(31 downto 0);
		signal   adr      : out std_logic_vector(31 downto 0);
		signal   we       : out std_logic;
		signal   cyc      : out std_logic;
	   	variable status   : out wb_status_type
	); 
	
	-- purpose: citanje podataka sa WB slava
	-- usage  : unutar procesa
	procedure wbs_rd(
		         data     : out  std_logic_vector;
		constant address  : in  std_logic_vector;
		constant data_len : in  positive;
		signal   clk      : in  std_logic;
		signal   stb      : out std_logic;
		signal   ack      : in  std_logic;
		signal   err      : in  std_logic;
		signal   rty      : in  std_logic;
		signal   sel      : out std_logic_vector( 3 downto 0);
		signal   dat      : in  std_logic_vector(31 downto 0);
		signal   adr      : out std_logic_vector(31 downto 0);
		signal   we       : out std_logic;
		signal   cyc      : out std_logic;
	   	variable status   : out wb_status_type
	);

	-- purpose: upis i citanje podataka u/iz WB slave
	--          podaci su pripremljeni u nizu data_array
	-- usage  : unutar procesa
	procedure wbs_testvector(
		variable data_array: inout wb_cycle_data_array;
		signal   clk      : in  std_logic;
		signal   stb      : out std_logic;
		signal   ack      : in  std_logic;
		signal   err      : in  std_logic;
		signal   rty      : in  std_logic;
		signal   sel      : out std_logic_vector( 3 downto 0);
		signal   dat_out  : out std_logic_vector(31 downto 0);
		signal   dat_in   : in  std_logic_vector(31 downto 0);
		signal   adr      : out std_logic_vector(31 downto 0);
		signal   we       : out std_logic;
		signal   cyc      : out std_logic
	);

	-- purpose: generisanje WB takta
    -- inputs : time - clock period
    -- outputs: clk
	-- usage  : kao konkurentni poziv procedure
	--          clk_gen: wb_clk_gen(wb_clk, 10ns);
	procedure wb_clk_gen(
		signal   clk      : out std_logic;
		constant t_clk    : in  time
	);	 
	 
	-- purpose: generisanje WB reset signala
    -- inputs : t_rst  - trajanje	  
	--          t_init - kasnjenje
    -- outputs: rst
	-- usage  : kao konkurentni poziv procedure
	--          clk_gen: wb_clk_gen(wb_clk, 10ns);
	procedure wb_rst_gen(
		signal   rst      : out std_logic;
		constant t_rst    : in  time;
		constant t_init   : in  time
	);
	
    -- purpose: generisanje ACK signala
    -- inputs : latency - broj Wait State stanja
    -- outputs: ack
	-- usage  : unutar procesa
	procedure wbs_gen_ack_response(
		signal   clk      : in  std_logic;
		signal   stb      : in  std_logic;
		signal   ack      : out std_logic;
		constant latency  : in  natural
	); 	 
	
    -- purpose: generisanje RTY signala
    -- outputs: rty
	-- usage  : unutar procesa
	procedure wbs_gen_rty_response(
		signal   clk      : in  std_logic;
		signal   stb      : in  std_logic;
		signal   rty      : out std_logic
	);
		 
    -- purpose: generisanje ERR signala
    -- outputs: err
	-- usage  : unutar procesa
	procedure wbs_gen_err_response(
		signal   clk      : in  std_logic;
		signal   stb      : in  std_logic;
		signal   err      : out std_logic
	);
		
---------------------------------------------------------------------- 
-- signal declarations 
---------------------------------------------------------------------- 
	-- WB common signals
	signal wb_clk         : std_logic := '0';
	signal wb_rst         : std_logic := '0';

	-- WB slave signals
	signal wbs_stb_i      : std_logic := '0';
	signal wbs_cyc_i      : std_logic := '0';
	signal wbs_we_i       : std_logic := '0';
	signal wbs_sel_i      : std_logic_vector( 3 downto 0) := "0000";	 
	signal wbs_dat_i      : std_logic_vector(31 downto 0) := (others => 'X');
	signal wbs_adr_i      : std_logic_vector(31 downto 0) := (others => 'X');
	signal wbs_ack_o      : std_logic := '0';
	signal wbs_err_o      : std_logic := '0'; 
	signal wbs_rty_o      : std_logic := '0';  
	signal wbs_dat_o      : std_logic_vector(31 downto 0) := (others => 'X');
		
	-- WB master signals
	signal wbm_stb_o      : std_logic := '0';
	signal wbm_cyc_o      : std_logic := '0';
	signal wbm_we_o       : std_logic := '0';
	signal wbm_sel_o      : std_logic_vector( 3 downto 0) := "0000";	 
	signal wbm_dat_o      : std_logic_vector(31 downto 0) := (others => 'X');
	signal wbm_adr_o      : std_logic_vector(31 downto 0) := (others => 'X');
	signal wbm_ack_i      : std_logic := '0';
	signal wbm_err_i      : std_logic := '0'; 
	signal wbm_rty_i      : std_logic := '0';  
	signal wbm_dat_i      : std_logic_vector(31 downto 0) := (others => 'X');
			
end package wb_test_suite;

package body wb_test_suite is
	
	constant Td: time := 2 ns;	
	 
	function log2(A: integer) return integer is
	begin
		for I in 1 to 30 loop   -- Works for up to 32 bit integers
			if(2**I > A) then
				return(I-1);
			end if;
		end loop;
		return(30);
 	end function log2;

	function vector_to_string (I: in std_logic_vector) return string is
    	variable output_string: string((I'length + I'length/4 + 1) downto 1); -- izlazni string
      	variable norm_index   : positive:=1;               -- index
	begin	
    	if (I'right > I'left) then						   -- ako je niz rastuci
        	norm_index := 1;							   -- pocni od pocetka
      	else                                               -- ako je opadajuci
         	norm_index := I'length + I'length/4 + 1;	   -- pocni od kraja
      	end if;

      	for index in I'range loop						   -- za sve elemente
			if( (index + 1) mod 4 = 0) then 
            	output_string(norm_index) := '_';
				if (I'right > I'left) then                      -- ako je niz rastuci
	            	norm_index := norm_index + 1;               -- uvecaj index	                 
	         	elsif (norm_index > 1) then						-- u suprotnom 
	            	norm_index := norm_index - 1;				-- umanji index
	         	end if;
			end if;
			
	       	if X01(I(index)) = '0' then								
            	output_string(norm_index) := '0';
         	elsif X01(I(index)) = '1' then
            	output_string(norm_index) := '1';
         	else
            	output_string(norm_index) := 'X';
         	end if; 
     
			if (I'right > I'left) then                      -- ako je niz rastuci
            	norm_index := norm_index + 1;               -- uvecaj index	                 
         	elsif (norm_index > 1) then						-- u suprotnom 
            	norm_index := norm_index - 1;				-- umanji index
         	end if;
      	end loop;			  
		return output_string;
	end function vector_to_string;	

	function get_sel(
		constant address     : in std_logic_vector(1 downto 0);
		constant data_length : in positive
	) return std_logic_vector is	
 		variable sel  : std_logic_vector (3 downto 0);
		variable size : integer;
		variable shift: integer;
	begin			   
		size := data_length / 8;
		
		if (size = 0) then
			size := 1;
		end if;
		
		sel(3 downto size)     := (others => '0');			
		sel(size - 1 downto 0) := (others => '1');
				
		shift := CONV_INTEGER(address);	

		if (shift > 0) then
			sel := To_StdLogicVector(To_bitvector(sel) sll shift);
		end if;

		return sel;
		
	end function get_sel;

	procedure wb_clk_gen(
		signal clk: out std_logic;
		constant t_clk: in time
	) is
	begin
		loop 
			clk	<= '0', '1' after t_clk/2;
			wait for t_clk;
		end loop;
	end procedure wb_clk_gen;	
	
	procedure wb_rst_gen(
		signal   rst: out std_logic;
		constant t_rst: in time;
		constant t_init: in time
	) is
	begin
		rst	<= '0';
		rst	<= '1' after t_init, '0' after t_init + t_rst;
		wait;
	end procedure wb_rst_gen;	

	procedure wbs_gen_ack_response(
		signal   clk      : in  std_logic;
		signal   stb      : in  std_logic;
		signal   ack      : out std_logic;
		constant latency  : in  natural
	) is		   
		variable cnt: natural;
	begin 
		ack <= '0' after Td;
		cnt := 0;
		
		if (stb = '0') then
			wait until rising_edge(stb);
		end if;

		if (latency /= 0) then
			lat: while (cnt < latency) loop	
				cnt := cnt + 1;	 
				wait until rising_edge(clk);
			end loop lat;
		end if;
		
		if (stb = '1') then
			ack <= '1' after Td;  
		end if;
									
		wait until (falling_edge(stb) or rising_edge(clk));

		ack <= '0' after Td;  
		wait for 0 ns;
	end procedure wbs_gen_ack_response;

	procedure wbs_gen_rty_response(
		signal   clk      : in  std_logic;
		signal   stb      : in  std_logic;
		signal   rty      : out std_logic
	) is		   
	begin 
		rty <= '0' after Td;   
		
		if (stb = '0') then
			wait until rising_edge(stb);
		end if;
		
		rty <= '1' after Td;  
		wait until rising_edge(clk);
		rty <= '0' after Td;  
		wait for 0 ns;
	end procedure wbs_gen_rty_response;

	procedure wbs_gen_err_response(
		signal   clk      : in  std_logic;
		signal   stb      : in  std_logic;
		signal   err      : out std_logic
	) is		   
	begin 
		err <= '0' after Td;   
		
		if (stb = '0') then
			wait until rising_edge(stb);
		end if;
		
		err <= '1' after Td;  
		wait until rising_edge(clk);
		err <= '0' after Td;  
		wait for 0 ns;
	end procedure wbs_gen_err_response;

	-- purpose: strobe generation
    -- inputs : ack, err, rty
    -- outputs: stb
	-- usage  : inside process
	procedure wbs_gen_stb_request(
		signal clk       : in  std_logic;
		signal stb       : out std_logic
	) is	
	begin
		stb <= '1' after Td; 
		wait for 0 ns;
	end procedure wbs_gen_stb_request;
	 
	procedure wbs_wait_stb_response (
		signal clk       : in  std_logic;
		signal stb       : out std_logic;
		signal ack       : in  std_logic;
		signal err       : in  std_logic;
		signal rty       : in  std_logic;
		variable status  : out wb_status_type
	)is
	begin 
		status := OK;
		wait_ack: loop 	  
			wait until rising_edge(clk);
			if ack = '1' then
				stb <= '0' after Td; 
				exit wait_ack;	
			elsif rty = '1' then 
				stb <= '0' after Td; 
				status := RETRY;
				exit wait_ack;	
			elsif err = '1' then 
				stb <= '0' after Td; 
				status := ERROR;
				exit wait_ack;	
			end if;
			
		end loop wait_ack;		 
		wait for 0 ns;
	end procedure wbs_wait_stb_response; 
	
	procedure wbs_wr(
		constant data   : in  std_logic_vector;
		constant address: in  std_logic_vector;
		signal   clk    : in  std_logic;
		signal   stb    : out std_logic;
		signal   ack    : in  std_logic;
		signal   err    : in  std_logic;
		signal   rty    : in  std_logic;
		signal   sel    : out std_logic_vector( 3 downto 0);
		signal   dat    : out std_logic_vector(31 downto 0); 
		signal   adr    : out std_logic_vector(31 downto 0);
		signal   we     : out std_logic;
		signal   cyc    : out std_logic;
	   	variable status : out wb_status_type
	) is
		variable adr_sel  : std_logic_vector(1 downto 0);  
		variable dat_sel  : std_logic_vector(3 downto 0); 
	begin 		   
		if (address'left > address'right) then
			adr_sel := address(1 downto 0);
		else
			adr_sel := address(address'right - 1) & address(address'right);
		end if;
		
		
		dat_sel := get_sel(
			adr_sel, 
			data'length
		);	 
			   
		sel <= dat_sel;

		we  <= '1' after Td;
		cyc <= '1' after Td;
		dat <= (others => '0') after Td;   
		
		case (dat_sel) is
			when "0001" | "0011" | "0111" | "1111" => 
				dat(data'length - 1 downto 0)       <= data after Td;
			when "0010" | "0110" | "1110"  => 
				dat(data'length - 1 + 8 downto 8)   <= data after Td;
			when "0100" | "1100" => 
				dat(data'length - 1 + 16 downto 16) <= data after Td;
			when "1000" =>
				dat(data'length - 1 + 24 downto 24) <= data after Td;
			when others => null;
		end case;
		
		adr <= (others => '0') after Td;
		adr(address'high downto address'low) <= address after Td;

		wbs_gen_stb_request(
			clk, 
			stb
		);	 
		
		wbs_wait_stb_response (
			clk, 
			stb, 
			ack, 
			err, 
			rty,
			status
		);
		 
		we  <= '0' after Td;
		cyc <= '0' after Td;
		wait for 0 ns;
	end procedure wbs_wr;	 
	
	procedure wbs_rd(
		         data     : out std_logic_vector;
		constant address  : in  std_logic_vector;
		constant data_len : in  positive;
		signal   clk      : in  std_logic;
		signal   stb      : out std_logic;
		signal   ack      : in  std_logic;
		signal   err      : in  std_logic;
		signal   rty      : in  std_logic;
		signal   sel      : out std_logic_vector( 3 downto 0);
		signal   dat      : in  std_logic_vector(31 downto 0);
		signal   adr      : out std_logic_vector(31 downto 0);
		signal   we       : out std_logic;
		signal   cyc      : out std_logic;
	   	variable status   : out wb_status_type
	) is
		variable dat_sel  : std_logic_vector(3 downto 0);  
		variable adr_sel  : std_logic_vector(1 downto 0);
	begin
		if (address'left > address'right) then
			adr_sel := address(1 downto 0);
		else
			adr_sel := address(address'right - 1) & address(address'right);
		end if;
		
		dat_sel := get_sel(
			adr_sel, 
			data_len
		);	 
			   
		sel <= dat_sel;

		we  <= '0' after Td;
		cyc <= '1' after Td; 

		data(data'range):= (others => '0');
		
		adr <= (others => '0') after Td;
		adr(address'high downto address'low) <= address after Td;

		wbs_gen_stb_request(
			clk, 
			stb
		);	 
		
		wbs_wait_stb_response (
			clk, 
			stb, 
			ack, 
			err, 
			rty,
			status
		);
		
		case (dat_sel) is
			when "0001" | "0011" | "0111" | "1111" => 
				-- cisti sve sem najlijevlje '1'
--					report "dat_sel prije: " & vector_to_string(dat_sel);
					dat_sel := To_StdLogicVector(To_bitvector(dat_sel) srl (data_len/8-1)); 
					dat_sel := To_StdLogicVector(To_bitvector(dat_sel) sll (data_len/8-1)); 
--					report "dat_sel poslje: " & vector_to_string(dat_sel);
			data(7*CONV_INTEGER(dat_sel) downto 0) := dat(7*CONV_INTEGER(dat_sel) downto 0);
--				report "Data je []: " & vector_to_string(dat(7*CONV_INTEGER(dat_sel) downto 0));
			when "0010" | "0110" | "1110"  =>
				dat_sel := To_StdLogicVector(To_bitvector(dat_sel) srl 1);
				-- cisti sve sem najlijevlje '1'
--					report "dat_sel prije: " & vector_to_string(dat_sel);
					dat_sel := To_StdLogicVector(To_bitvector(dat_sel) srl (data_len/8-1)); 
					dat_sel := To_StdLogicVector(To_bitvector(dat_sel) sll (data_len/8-1)); 
--					report "dat_sel poslje: " & vector_to_string(dat_sel);
				data(7*CONV_INTEGER(dat_sel) downto 0) := dat(7*CONV_INTEGER(dat_sel)+ 8 downto 8);
--				report "Data je [0]: " & vector_to_string(dat(7*CONV_INTEGER(dat_sel)+ 16 downto 16));
			when "0100" | "1100" => 
				dat_sel := To_StdLogicVector(To_bitvector(dat_sel) srl 2);
				-- cisti sve sem najlijevlje '1'
					dat_sel := To_StdLogicVector(To_bitvector(dat_sel) srl (data_len-1)); 
					dat_sel := To_StdLogicVector(To_bitvector(dat_sel) sll (data_len-1)); 
				data(7*CONV_INTEGER(dat_sel) downto 0) := dat(7*CONV_INTEGER(dat_sel)+ 16 downto 16);
--				report "Data je [00]: " & vector_to_string(dat(7*CONV_INTEGER(dat_sel)+ 16 downto 16));
			when "1000" =>
				dat_sel := To_StdLogicVector(To_bitvector(dat_sel) srl 3);
				-- cisti sve sem najlijevlje '1'
					dat_sel := To_StdLogicVector(To_bitvector(dat_sel) srl (data_len-1)); 
					dat_sel := To_StdLogicVector(To_bitvector(dat_sel) sll (data_len-1)); 
				data(7*CONV_INTEGER(dat_sel) downto 0) := dat(7*CONV_INTEGER(dat_sel)+ 24 downto 24);
--				report "Data je[000]: " & vector_to_string(dat(7*CONV_INTEGER(dat_sel)+ 24 downto 24));
			when others => null;
		end case;

		we  <= '0' after Td;
		cyc <= '0' after Td;
	end procedure wbs_rd;	

	procedure wbs_testvector(
		variable data_array: inout wb_cycle_data_array;
		signal   clk      : in  std_logic;
		signal   stb      : out std_logic;
		signal   ack      : in  std_logic;
		signal   err      : in  std_logic;
		signal   rty      : in  std_logic;
		signal   sel      : out std_logic_vector( 3 downto 0);
		signal   dat_out  : out std_logic_vector(31 downto 0);
		signal   dat_in   : in  std_logic_vector(31 downto 0);
		signal   adr      : out std_logic_vector(31 downto 0);
		signal   we       : out std_logic;
		signal   cyc      : out std_logic
	) is   
		variable data: std_logic_vector(31 downto 0);  
		variable first_time: boolean:= true;
		variable last_wr   : boolean:= true;   
		variable wb_status : wb_status_type;
	begin  
		for index in data_array'range loop 	
			wb_status := OK;
			if data_array(index).wr then

				if (not first_time and not last_wr) then
					wait until rising_edge(clk);
				end if;
				
				WAIT_WR_ACK: loop
				wbs_wr(
					data_array(index).value(data_array(index).data_width - 1 downto 0),
					data_array(index).address(data_array(index).adr_width - 1 downto 0),
					clk,
					stb,
					ack,
					err,
					rty,
					sel,
					dat_out,
					adr,
					we,
					cyc,
					wb_status
				);	   
					exit WAIT_WR_ACK when wb_status /= RETRY; 
				end loop WAIT_WR_ACK;
				first_time:= false;
				last_wr := true;
			else 
				if (not first_time and last_wr) then
					wait until rising_edge(clk);
				end if;
				
				WAIT_RD_ACK: loop
				wbs_rd(
					data_array(index).value,
					data_array(index).address(data_array(index).adr_width - 1 downto 0),
					data_array(index).data_width,
					clk,
					stb,
					ack,
					err,
					rty,
					sel,
					dat_in,
					adr,
					we,
					cyc,
					wb_status
				); 
				
				if (wb_status = OK) then
					report "Ocekivano: " & vector_to_string(data_array(index).exp_value);
					report "Procitano: " & vector_to_string(data_array(index).value);
				end if;

					exit WAIT_RD_ACK when wb_status /= RETRY; 
				end loop WAIT_RD_ACK;
				   
				first_time:= false;
				last_wr := false;
			end if;
		end loop;  
	end procedure wbs_testvector;
		
end package body wb_test_suite;



















